/* esdata.cpp */

#include "esdata.h"
#include "utilio.h"
#include <expat.h>
#include <stdio.h>
#include <malloc.h>	//free() for strdup()

UnitGroupLink::UnitGroupLink()
:	list(NULL)
{
}

UnitGroupLink::~UnitGroupLink()
{
	delete [] list;
}

CivLink::CivLink()
:	units(NULL), unit_tail(NULL)
{
}

CivLink::~CivLink()
{
	destroylist(units);
}

ESDATA::ESDATA()
:	techs(NULL), tech_tail(NULL),
	colors(NULL), color_tail(NULL),
	units(NULL), unit_tail(NULL),
	resources(NULL), res_tail(NULL),
	unitgroups(NULL), unitgroup_tail(NULL),
	ug_buildings(NULL), ug_units(NULL)
{
}

ESDATA::~ESDATA()
{
	/*
	Link *parse = techs;

	while (parse)
	{
		techs = (TechLink*)parse;
		parse = parse->next;

		delete techs;
	}

	parse = colors;
	while (parse)
	{
		colors = (ColorLink*)parse;
		parse = parse->next;

		delete colors;
	}*/
	destroylist(techs);
	destroylist(colors);
}

void ESDATA::readTech(const XML_Char **attrs)
{
	TechLink *link = new TechLink;

	while (*attrs)
	{
		if (link->read(*attrs, attrs[1]))
			attrs++;
		else
			attrs++;

		attrs++;
	}

	addNode((Link*&)techs, (Link*&)tech_tail, link);
	tech_count++;
}

void ESDATA::readColor(const XML_Char **attrs)
{
	ColorLink *link = new ColorLink;

	while (*attrs)
	{
		if (link->read(*attrs, attrs[1]))
			attrs++;
		else if (!strcmp(*attrs, "color"))
			sscanf(*++attrs, "%x", &link->ref);
		else
			attrs++;

		attrs++;
	}

	addNode((Link*&)colors, (Link*&)color_tail, link);
	color_count++;
}

void ESDATA::readUnit(const XML_Char **attrs)
{
	UnitLink *link = new UnitLink;

	while (*attrs)
	{
		if (link->read(*attrs, attrs[1]))
			attrs++;
		else
			attrs++;

		attrs++;
	}

	addNode((Link*&)units, (Link*&)unit_tail, link);
	unit_count++;
}

void ESDATA::readTerrain(const XML_Char **attrs)
{
	ColorLink *link = new ColorLink;

	while (*attrs)
	{
		if (link->read(*attrs, attrs[1]))
			attrs++;
		else if (!strcmp(*attrs, "color"))
			sscanf(*++attrs, "%x", &link->ref);
		else
			attrs++;

		attrs++;
	}

	addNode((Link*&)terrains, (Link*&)terrain_tail, link);
	terrain_count++;
}

void ESDATA::readSimple(const XML_Char **attrs)
{
	Link *link = new Link;
	Link **dest = NULL, **dest_tail = NULL;

	while (*attrs)
	{
		if (link->read(*attrs, attrs[1]))
			attrs++;
		else
			attrs++;

		attrs++;
	}

	switch (rstate.pos)
	{
	case ESD_resources:
		dest = &resources;
		dest_tail = &res_tail;
		res_count++;
		break;
	case ESD_aitypes:
		dest = &aitypes;
		dest_tail = &aitype_tail;
		aitype_count++;
		break;
	}

	if (dest && dest_tail)
		addNode(*dest, *dest_tail, link);
}

void ESDATA::readResources(const XML_Char **attrs)
{
	rstate.pos = ESD_resources;
}

void ESDATA::readAitypes(const XML_Char **attrs)
{
	rstate.pos = ESD_aitypes;
}

void ESDATA::readUnitGroup(const XML_Char **attrs)
{
	UnitGroupLink *link = new UnitGroupLink;
	const UnitLink **uparse;
	char *dup, *p;
	int i;

	while (*attrs)
	{
		if (link->read(*attrs, attrs[1]))
			attrs++;
		else if (!strcmp(*attrs, "units"))
		{
			dup = strdup(attrs[1]);

			/* Count the number of units in group */
			p = dup;
			link->count = 1;	//one more unit than commas
			while (p = strchr(p, ','))
			{
				link->count++;
				p++;	//skip past comma
			}

			/* Allocate and load the list */
			link->list = new const UnitLink*[link->count];
			uparse = link->list;
			if (!uparse)
				return;
			p = strtok(dup, ",");
			do
			{
				i = atoi(p);
				if (*uparse = (UnitLink*)getById(esdata.units, i))
					uparse++;
				else
					printf("Discarding non-existent unit type %d.\n", i);
			} while (p = strtok(NULL, ","));

			attrs++;
		}
		else
			attrs++;

		attrs++;
	}

	if (!link->name)
	{
		puts("ERROR: No name value for unit group link. Chucking.");
		return;
	}

	if (!link->list || !*link->list)
	{
		printf("Discarding empty unit group %s.\n", link->name);
		delete [] link->list;
		link->list = NULL;
		delete link;
		return;
	}

	if (!ug_buildings && !strcmp(link->name, "Buildings"))
		ug_buildings = link;
	else if (!ug_units && !strcmp(link->name, "Units"))
		ug_units = link;

	addNode((Link*&)unitgroups, (Link*&)unitgroup_tail, link);
	unitgroup_count++;
}

void ESDATA::readCiv(const XML_Char **attrs)
{
	CivLink *link = new CivLink;

	while (*attrs)
	{
		if (link->read(*attrs, attrs[1]))
			attrs++;
		else
			attrs++;

		attrs++;
	}

	addNode((Link*&)civs, (Link*&)civ_tail, link);
	civ_count++;

	rstate.civ = link;
}

#define PF(n, f) { n, f }	//I just prefer this syntax
ESDATA::pFunc ESDATA::pfuncs[] = 
{
	PF("tech", &ESDATA::readTech),
	PF("color", &ESDATA::readColor),
	PF("protounit", &ESDATA::readUnit),
	PF("resources", &ESDATA::readResources),
	PF("simple", &ESDATA::readSimple),
	PF("aitypes", &ESDATA::readAitypes),
	PF("terrain", &ESDATA::readTerrain),
	PF("group", &ESDATA::readUnitGroup),
	PF("civ", &ESDATA::readCiv),
	PF(NULL, NULL)
};

void XMLCALL startHandler(void *userdata, const XML_Char *name, const XML_Char **attrs)
{
	ESDATA *esdata = (ESDATA*)userdata;
	ESDATA::pFunc *pf = esdata->pfuncs;

	while (pf->name)
	{
		if (!strcmp(name, pf->name))
			break;
		pf++;
	}

/*	Now I know you'll look at this and think "omfg this looks so
	complex!", but it's really one of those things that looks more
	complicated to the coder than to the machine. The ASM is
	surprisingly simple. It's really much simpler than a bunch of
	if statements. */
	if (pf->name)
		(esdata->*(pf->f))(attrs);
}

/* End tags */

void ESDATA::endResources()
{
	rstate.pos = ESD_nowhere;
}

void ESDATA::endCiv()
{
	rstate.civ = NULL;
}

void XMLCALL endHandler(void *userdata, const XML_Char *name)
{
	ESDATA *esdata = (ESDATA *)userdata;

	if (!strcmp(name, "resources"))
		esdata->endResources();
}

void ESDATA::addNode(Link *&head, Link *&tail, Link *item)
{
	if (!head)
		head = item;

	if (tail)
		tail->next = item;

	tail = item;
}

bool ESDATA::load(const char *path)
{
	int size;
	XML_Parser parser;
	FILE *datafile;
	void *buffer;

	/* init */
	tech_count = 0;
	color_count = 0;
	unit_count = 0;
	res_count = 0;
	aitype_count = 0;
	terrain_count = 0;
	unitgroup_count = 0;
	civ_count = 0;
	rstate.pos = ESD_nowhere;
	rstate.civ = NULL;
	printf("\nReading esdata...\n");

	parser = XML_ParserCreate(NULL);
	if (!parser)
		return false;

	XML_SetElementHandler(parser, startHandler, endHandler);
	XML_SetUserData(parser, this);

	size = fsize(path);

	buffer = XML_GetBuffer(parser, size);
	datafile = fopen(path, "r");
	if (datafile)
	{
		fread(buffer, size, 1, datafile);
		fclose(datafile);
		XML_ParseBuffer(parser, size, 1);

		printf("Loaded:\n"
			"%d technologies,\n"
			"%d colors,\n"
			"%d units,\n"
			"%d resources,\n"
			"%d aitypes,\n"
			"%d terrains,\n"
			"%d unitgroups,\n"
			"and %d civilizations.",
			tech_count, color_count, unit_count, res_count, aitype_count,
			terrain_count, unitgroup_count, civ_count);
	}

	XML_ParserFree(parser);

	if (datafile)
		printf("Done.\n\n");
	else
		printf("Failed.\n\n");

	return (datafile != NULL);
}

int ESDATA::getCount(enum ESD_GROUP group)
{
	switch (group)
	{
	case ESD_techs:
		return tech_count;
	case ESD_units:
		return unit_count;
	case ESD_colors:
		return color_count;
	case ESD_resources:
		return res_count;
	case ESD_aitypes:
		return aitype_count;
	case ESD_terrains:
		return terrain_count;
	}

	return 0;
}

UnitLink *ESDATA::getUnitById(int id)
{
	UnitLink *ret;
	char buffer[20];

	ret = (UnitLink*)getById(esdata.units, id);

	if (!ret)
	{
		_snprintf(buffer, sizeof(buffer), "?%d", id);

		printf("Creating new protounit for unrecognized typeid %d.\n", id);

		ret = new UnitLink;
		ret->id = id;
		ret->name = strdup(buffer);

		addNode((Link*&)units, (Link*&)unit_tail, ret);
	}

	return ret;
}
