/* esdata.h: Manages standard AOE data. */

#include <expat.h>
#include "datatypes.h"

/* geniedata type links */

class TechLink : public Link
{
public:
	char *civ;
};

class ColorLink : public Link
{
public:
	long ref;
};

class UnitLink : public Link
{
public:
	long hp;
};

class UnitGroupLink : public Link
{
public:
	UnitGroupLink();
	~UnitGroupLink();

	unsigned count;
	const UnitLink **list;
};

class CivLink : public Link
{
public:
	CivLink();
	~CivLink();

	UnitLink *units, *unit_tail;
};

enum ESD_GROUP
{
	ESD_nowhere,
	ESD_techs,
	ESD_units,
	ESD_colors,
	ESD_resources,
	ESD_aitypes,
	ESD_terrains,
	ESD_unitgroups,
	ESD_civs
};

extern class ESDATA
{
	int tech_count, color_count, unit_count, res_count, aitype_count,
		terrain_count, unitgroup_count, civ_count;

	/*	I know, I should probably use a hashtable for this,
		but I don't expect adding enough different tags to
		warrant it. */
	static struct pFunc
	{
		char *name;
		void (ESDATA::*f)(const XML_Char **attrs);
	} pfuncs[];

	/* Reading funcs */
	void addNode(Link *&head, Link *&tail, Link *item);
	void readTech(const XML_Char **attrs);
	void readColor(const XML_Char **attrs);
	void readUnit(const XML_Char **attrs);
	void readTerrain(const XML_Char **attrs);
	void readSimple(const XML_Char **attrs);
	void readResources(const XML_Char **attrs);
	void readAitypes(const XML_Char **attrs);
	void readUnitGroup(const XML_Char **attrs);
	void readCiv(const XML_Char **attrs);

	void endResources();
	void endCiv();

public:
	/* LL heads */
	TechLink *techs;
	ColorLink *colors;
	UnitLink *units;
	Link *resources;
	Link *aitypes;
	ColorLink *terrains;
	UnitGroupLink *unitgroups;
	CivLink *civs;

	/* Special unit groups */
	UnitGroupLink *ug_buildings, *ug_units;

	ESDATA();
	~ESDATA();
	bool load(const char *path);

	int getCount(enum ESD_GROUP group);
	UnitLink *getUnitById(int id);

private:
	/* LL Tails */
	TechLink *tech_tail;
	ColorLink *color_tail;
	UnitLink *unit_tail;
	Link *res_tail;
	Link *aitype_tail;
	ColorLink *terrain_tail;
	UnitGroupLink *unitgroup_tail;
	CivLink *civ_tail;

	struct
	{
		enum ESD_GROUP pos;
		CivLink *civ;
	} rstate;

	friend void XMLCALL startHandler(void *, const XML_Char *, const XML_Char **);
	friend void XMLCALL endHandler(void *, const XML_Char *);
} esdata;

