/**
	AOK Trigger Studio (See aokts.cpp for legal conditions.)
	WINDOWS VERSION
	unitedit.cpp -- Functions for the unit editor.
**/

#include "editors.h"
#include <commctrl.h>
#include <stdio.h>
#include "scen.h"
#include "settings.h"
#include "resource.h"
#include "utilui.h"
#include "utilunit.h"
#include "mapview.h"

index_t u_index = (unsigned)-1;	//offset of currently selected unit in vector
unsigned int c_index = (unsigned)-1;	//LB index of currently selected constant

#define NUM_LISTS	5
#define NUM_ROTATES 8
const char *rotates[NUM_ROTATES] = { "0/4", "1/4", "2/4", "3/4", "4/4", "5/4", "6/4", "7/4" };

const char szTitle[] =
"Unit Editor";
const char errorUnfoundUnit[] =
"Could not find unit to load. I most likely incorrectly deleted a unit.";
const char errorUnfoundType[] =
"You have some strange units on this map, my friend. Unpredictable results may occur.";
const char warningNoSelDelete[] =
"Unknown unit selected! Deleting aborted.";

int UnitList_FindUnit(HWND unitbox, const UnitLink *type)
{
	int index, ret = -1;

	index = SendMessage(unitbox, LB_GETCOUNT, 0, 0);

	while (index--)
	{
		if (SendMessage(unitbox, LB_GETITEMDATA, index, 0) == (int)type)
		{
			ret = index;
			break;
		}
	}

	return ret;
}

bool Units_Load(HWND dialog)
{
	bool ret = true;
	Unit *u;
	Unit blank;	//for dummy purposes
	HWND unitbox = GetDlgItem(dialog, IDC_U_UNIT);
	const UnitLink *type;

	if (u_index == -1)
	{
		u = &blank;
		u->setType(esdata.units);
	}

	else if ((signed)u_index < propdata.p->units.count())
	{
		u = propdata.p->units.at(u_index);
		SendMessage(mapview, MAP_HighlightPoint, (WPARAM)u->x, (LPARAM)u->y);
	}

	else
	{
		u = &blank;
		ret = false;
		MessageBox(dialog, errorUnfoundUnit, szTitle, MB_ICONERROR);
	}

	if (!(type = u->getType()))
	{
		MessageBox(dialog, errorUnfoundType, szTitle, MB_ICONERROR);
		return false;
	}
	SetDlgItemInt(dialog, IDC_U_CONST, type->id, FALSE);
	c_index = UnitList_FindUnit(unitbox, type);

	//if unit is not in current list, change list to "All"
	if (c_index == -1)
	{
		SendDlgItemMessage(dialog, IDC_U_TYPE, CB_SETCURSEL, 0, 0);
		UnitList_FillGroup(unitbox, NULL);

		//now search for unit on the "All" list. (If this fails, too bad.)
		if ((c_index = UnitList_FindUnit(unitbox, type)) == -1)
			MessageBox(dialog, errorUnfoundType, szTitle, MB_ICONWARNING);
	}

	SendDlgItemMessage(dialog, IDC_U_UNIT, LB_SETCURSEL, c_index, 0);

	SetDlgItemFloat(dialog, IDC_U_X, u->x);
	SetDlgItemFloat(dialog, IDC_U_Y, u->y);
	SendDlgItemMessage(dialog, IDC_U_ROTATE, CB_SETCURSEL, (int)(u->rotate / PI * 4), 0);
	SetDlgItemInt(dialog, IDC_U_FRAME, u->frame, FALSE);
	SetDlgItemInt(dialog, IDC_U_GARRISON, u->garrison, TRUE);

	return ret;
}

void Units_Save(HWND dialog)
{
	Unit *u = NULL;
	HWND control;

	if (u_index != LB_ERR && (u = propdata.p->units.at(u_index)))
	{
		u->x =		GetDlgItemFloat(dialog, IDC_U_X);
		u->y =		GetDlgItemFloat(dialog, IDC_U_Y);
		u->setType((UnitLink*)GETLBSELDATA(IDC_U_UNIT));
		u->rotate =	(float)SendDlgItemMessage(dialog, IDC_U_ROTATE, CB_GETCURSEL, 0, 0) / 4 * (float)PI;
		u->frame = (short)GetDlgItemInt(dialog, IDC_U_FRAME, NULL, FALSE);
		u->garrison = GetDlgItemInt(dialog, IDC_U_GARRISON, NULL, TRUE);
	}
}

/*
	UnitList_ChangeType: Handles change the type constant of a unit.
*/
void UnitList_ChangeType(HWND dialog, HWND typebox)
{
	unsigned int type_index;
	unsigned int selected_index;	//index of selected unit in listbox;
	HWND selbox, ttext;
	SVector <Unit> *units;
	enum Sorts sort;
	char string[10];
	UnitLink *type;

	selbox = GetDlgItem(dialog, IDC_U_SELU);
	ttext  = GetDlgItem(dialog, IDC_U_CONST);
	units = &propdata.p->units;
	sort = (enum Sorts)SendDlgItemMessage(dialog, IDC_U_SORT, CB_GETCURSEL, 0, 0);

	type_index = SendMessage(typebox, LB_GETCURSEL, 0, 0);

	if (type_index != c_index && u_index != -1)	//not the constant we had selected before
	{
		selected_index = SendMessage(selbox, LB_GETCURSEL, 0, 0);

		Unit *u = units->at(u_index);

		Units_Save(dialog);

		//then "update" the string in the selection box
		SendMessage(selbox, LB_DELETESTRING, selected_index, 0);
		selected_index = UnitList_InsertItem(selbox, sort, units->first(), units->count(), u);
		SendMessage(selbox, LB_SETCURSEL, selected_index, 0);
	}

	c_index = type_index;	//update this even if no unit selected
	type = (UnitLink*)SendMessage(typebox, LB_GETITEMDATA, type_index, 0);
	if (type)
		SetWindowText(ttext, itoa(type->id, string, 10));
}

void Units_Reset(HWND dialog)
{
	enum Sorts sort = (enum Sorts)SendDlgItemMessage(dialog, IDC_U_SORT, CB_GETCURSEL, 0, 0);
	UnitList_Fill(
		GetDlgItem(dialog, IDC_U_SELU),
		sort,
		propdata.p->units.first(),
		propdata.p->units.count()
		);
	u_index = (unsigned)-1;
	ENABLE_WND(IDC_U_DEL, false);
	Units_Load(dialog);
}

void Units_HandleDelete(HWND dialog)
{
	unsigned index, count, data;
	HWND selbox = GetDlgItem(dialog, IDC_U_SELU);

	if (u_index == -1)
	{
		MessageBox(dialog, warningNoSelDelete, szTitle, MB_ICONWARNING);
		return;
	}

	propdata.p->units.remove(u_index);

	index = SendMessage(selbox, LB_GETCURSEL, 0, 0);
	SendMessage(selbox, LB_DELETESTRING, index, 0);

	/* Update the other items' data. */
	count = SendMessage(selbox, LB_GETCOUNT, 0, 0);
	for (unsigned i = 0; i < count; i++)
	{
		data = SendMessage(selbox, LB_GETITEMDATA, i, 0);
		if (data > u_index)
			SendMessage(selbox, LB_SETITEMDATA, i, data - 1);
	}

	/* Set selection to next item down. If there isn't one, reset stuff. */
	if (SendMessage(selbox, LB_GETCOUNT, 0, 0))
	{
		SendMessage(selbox, LB_SETCURSEL, index, 0);
		u_index = SendMessage(selbox, LB_GETITEMDATA, index, 0);
		Units_Load(dialog);
	}
	else
	{
		ENABLE_WND(IDC_U_DEL, false);
		u_index = (unsigned)-1;
	}
}

void Units_HandleSelChange(HWND dialog, HWND listbox)
{
	int index;

	Units_Save(dialog);
	SendMessage(mapview, MAP_UnhighlightPoint, MAP_UNHIGHLIGHT_ALL, 0);
	index = SendMessage(listbox, LB_GETCURSEL, 0, 0);
	u_index = SendMessage(listbox, LB_GETITEMDATA, index, 0);

	if (u_index >= 0)
	{
		EnableWindow(GetDlgItem(dialog, IDC_U_DEL), TRUE);
		EnableMenuItem(propdata.menu, ID_EDIT_DELETE, MF_ENABLED);
	}

	Units_Load(dialog);
}

void Units_HandleTypeChange(HWND dialog, HWND typelist)
{
	int index = SendMessage(typelist, CB_GETCURSEL, 0, 0);
	HWND unitlist = GetDlgItem(dialog, IDC_U_UNIT);

	UnitList_FillGroup(unitlist,
		(UnitGroupLink *)SendMessage(typelist, CB_GETITEMDATA, index, 0));
}

void Units_HandleAdd(HWND dialog)
{
	Unit u(scen.next_uid++);
	char string[0x30], name[32];
	int index;
	HWND listbox = GetDlgItem(dialog, IDC_U_SELU);
	HWND typebox = GetDlgItem(dialog, IDC_U_UNIT);

	u.setType((UnitLink*)SendMessage(typebox, LB_GETITEMDATA, c_index, 0));
	SendMessage(typebox, LB_GETTEXT, c_index, (LPARAM)name);				
	u_index = propdata.p->units.append(u);

	sprintf(string, "%d: %s", u.ident, name);
	index = SendMessage(listbox, LB_ADDSTRING, 0, (LPARAM)string);

	// Select new item.
	SendMessage(listbox, LB_SETITEMDATA, index, u_index);
	SendMessage(listbox, LB_SETCURSEL, index, 0);
	ENABLE_WND(IDC_U_DEL, true);
}

void Units_HandleCommand(HWND dialog, WORD code, WORD id, HWND control)
{
	switch (code)
	{
	case BN_CLICKED:	//and menuitem
	case CBN_SELCHANGE:	//and accelerator, LBN_SELCHANGE
		switch (id)
		{
		case IDC_U_SELP:	//CBN_SELCHANGE
			Units_Save(dialog);
			propdata.pindex = SendMessage(control, CB_GETCURSEL, 0, 0);
			propdata.p = scen.players + propdata.pindex;
			Units_Reset(dialog);
			break;

		case IDC_U_SELU:	//LBN_SELCHANGE
			Units_HandleSelChange(dialog, control);
			break;

		case IDC_U_TYPE:	//CBN_SELCHANGE
			Units_HandleTypeChange(dialog, control);
			break;

		case IDC_U_ADD:		//BN_CLICKED
			Units_HandleAdd(dialog);
			break;

		case IDC_U_DEL:			//BN_CLICKED
		case ID_EDIT_DELETE:	//accelerator
			Units_HandleDelete(dialog);
			break;

		case IDC_U_UNIT:		//LBN_SELCHANGE
			UnitList_ChangeType(dialog, control);
			break;

		case IDC_U_SORT:
			Units_Reset(dialog);
			break;

		case ID_EDIT_COPY:
			SendMessage(GetFocus(), WM_COPY, 0, 0);
			break;

		case ID_EDIT_CUT:
			SendMessage(GetFocus(), WM_CUT, 0, 0);
			break;

		case ID_EDIT_PASTE:
			SendMessage(GetFocus(), WM_PASTE, 0, 0);
			break;
		}
		break;

	case EN_SETFOCUS:
		EnableMenuItem(propdata.menu, ID_EDIT_COPY, MF_ENABLED);
		EnableMenuItem(propdata.menu, ID_EDIT_CUT, MF_ENABLED);
		if (IsClipboardFormatAvailable(CF_TEXT))
			EnableMenuItem(propdata.menu, ID_EDIT_PASTE, MF_ENABLED);
		break;

	case EN_KILLFOCUS:
		EnableMenuItem(propdata.menu, ID_EDIT_COPY, MF_GRAYED);
		EnableMenuItem(propdata.menu, ID_EDIT_CUT, MF_GRAYED);
		EnableMenuItem(propdata.menu, ID_EDIT_PASTE, MF_GRAYED);
		break;
	}
}

BOOL Units_HandleInit(HWND dialog)
{
	int index;
	unsigned int i;
	HWND control;
	Link *g_parse;

	FILL_CB(IDC_U_SELP, Player::names, NUM_PLAYERS);
	FILL_CB(IDC_U_ROTATE, rotates, NUM_ROTATES);
	FILL_CB(IDC_U_SORT, sorts, NUM_SORTS);
	SendMessage(control, CB_SETCURSEL, 0, 0);

	/* Fill Unit group list */
	control = GetDlgItem(dialog, IDC_U_TYPE);
	SendMessage(control, CB_ADDSTRING, 0, (LPARAM)"All");	//default all list
	SendMessage(control, CB_SETITEMDATA, 0, 0);
	for (g_parse = esdata.unitgroups; g_parse; g_parse = g_parse->next)
	{
		index = SendMessage(control, CB_ADDSTRING, 0, (LPARAM)g_parse->name);
		SendMessage(control, CB_SETITEMDATA, index, (LPARAM)g_parse);
	}
	SendMessage(control, CB_SETCURSEL, 0, 0);
	UnitList_FillGroup(GetDlgItem(dialog, IDC_U_UNIT), NULL);

	//set edit control limits
	SendDlgItemMessage(dialog, IDC_U_X, EM_SETLIMITTEXT, 5, 0);
	SendDlgItemMessage(dialog, IDC_U_Y, EM_SETLIMITTEXT, 5, 0);

	return TRUE;
}

void Units_HandleMapClick(HWND dialog, int x, int y)
{
	Unit *u;
	int ret;

	if (u_index >= 0)
	{
		u = propdata.p->units.at(u_index);
		ret = MessageBox(dialog,
			"Do you want to move the selected unit?",
			"Unit Editor", MB_YESNO);
		if (ret = IDYES)
		{
			SendMessage(mapview, MAP_UnhighlightPoint, u->x, u->y);
			SetDlgItemInt(dialog, IDC_U_X, x, FALSE);
			SetDlgItemInt(dialog, IDC_U_Y, y, FALSE);
			Units_Save(dialog);
			SendMessage(mapview, MAP_HighlightPoint, (WPARAM)x, (LPARAM)y);
			SendMessage(mapview, MAP_Reset, 0, 0);
		}
	}
}

INT_PTR CALLBACK UnitDlgProc(HWND dialog, UINT msg, WPARAM wParam, LPARAM lParam)
{
	INT_PTR ret = FALSE;

	switch (msg)
	{
	case WM_INITDIALOG:
		ret = Units_HandleInit(dialog);
		break;

	case WM_COMMAND:
		ret = 0;
		Units_HandleCommand(dialog, HIWORD(wParam), LOWORD(wParam), (HWND)lParam);
		break;

	case WM_NOTIFY:
		{
			NMHDR *header = (NMHDR*)lParam;
			switch (header->code)
			{
			case PSN_SETACTIVE:
				SendDlgItemMessage(dialog, IDC_U_SELP, CB_SETCURSEL, propdata.pindex, 0);
				Units_Reset(dialog);
				break;

			case PSN_KILLACTIVE:
				Units_Save(dialog);
				SendMessage(mapview, MAP_UnhighlightPoint, MAP_UNHIGHLIGHT_ALL, 0);
			}
		}
		break;

	case WM_VKEYTOITEM:
		if (LOWORD(wParam) == VK_DELETE)
			Units_HandleDelete(dialog);
		ret = -1;
		break;

	case AOKTS_Loading:
		Units_Reset(dialog);
		break;

	case AOKTS_Saving:
		Units_Save(dialog);
		break;

	case MAP_Click:
		Units_HandleMapClick(dialog, LOWORD(lParam), HIWORD(lParam));
		break;
	}

	return ret;
}
