/**
	AOK Trigger Studio (See aokts.cpp for legal conditions.)
	WINDOWS VERSION
	utilui.cpp -- Definitions of Utility Functions for UI
**/

#define WINVER 0x0400
#define _WIN32_IE 0x0300
#include <windows.h>
#include <commctrl.h>	//must be included before "utilui.h"

#include "datatypes.h"
#include "scen_const.h"
#include "utilui.h"
#include "resource.h"	//for LCOMBO

#include <stdio.h>
#include <stdlib.h>
#include <commdlg.h>

/* Globals */
HWND tooltip;	//set with TooltipInit()

/* LCOMBO */

WNDPROC ComboWndProc = NULL;
WNDPROC EditWndProc = NULL;

/* internal flags */
enum LCB_FLAGS
{
	LCBF_ALLOWNOSELECT = 0x01,
};

const char *noselect = "<none>";

int OnCBL_FILL(HWND combobox, Link *list, Link *select)
{
	int index, ret = -1;
	LONG_PTR userdata;

	userdata = GetWindowLongPtr(combobox, GWLP_USERDATA);

	SendMessage(combobox, CB_RESETCONTENT, 0, 0);

	if (userdata & LCBF_ALLOWNOSELECT)
	{
		index = SendMessage(combobox, CB_ADDSTRING, 0, (LPARAM)noselect);
		SendMessage(combobox, CB_SETITEMDATA, index, NULL);
	}

	for (; list; list = list->next)
	{
		index = SendMessage(combobox, CB_ADDSTRING, 0, (LPARAM)list->name);
		SendMessage(combobox, CB_SETITEMDATA, index, (LPARAM)list);
		if (list == select)
		{
			SendMessage(combobox, CB_SETCURSEL, index, 0);
			ret = index;
		}
	}

	if (ret == -1 && (userdata & LCBF_ALLOWNOSELECT))
	{
		SendMessage(combobox, CB_SETCURSEL, 0, 0);
		ret = 0;
	}

	return ret;
}

int OnCBL_FILLBYID(HWND combobox, Link *list, int select)
{
	int index, ret = -1;
	LONG_PTR userdata;

	userdata = GetWindowLongPtr(combobox, GWLP_USERDATA);

	SendMessage(combobox, CB_RESETCONTENT, 0, 0);

	if (userdata & LCBF_ALLOWNOSELECT)
	{
		index = SendMessage(combobox, CB_ADDSTRING, 0, (LPARAM)noselect);
		SendMessage(combobox, CB_SETITEMDATA, index, NULL);
	}

	for (; list; list = list->next)
	{
		index = SendMessage(combobox, CB_ADDSTRING, 0, (LPARAM)list->name);
		SendMessage(combobox, CB_SETITEMDATA, index, (LPARAM)list);
		if (list->id == select)
		{
			SendMessage(combobox, CB_SETCURSEL, index, 0);
			ret = index;
		}
	}

	if (ret == -1 && (userdata & LCBF_ALLOWNOSELECT))
	{
		SendMessage(combobox, CB_SETCURSEL, 0, 0);
		ret = 0;
	}

	return ret;
}

long OnCBL_GETSELID(HWND combobox)
{
	int index;
	long ret = -1;

	index = SendMessage(combobox, CB_GETCURSEL, 0, 0);
	if (index >= 0)
		ret = ((Link*)SendMessage(combobox, CB_GETITEMDATA, index, 0))->id;

	return ret;
}

Link *OnCBL_GETSELPTR(HWND combobox)
{
	int index;
	Link *ret = NULL;

	index = SendMessage(combobox, CB_GETCURSEL, 0, 0);
	ret = (Link*)SendMessage(combobox, CB_GETITEMDATA, index, 0);

	return ((int)ret != LB_ERR) ? ret : NULL;
}

int OnCBL_FIND(HWND combobox, LPARAM data)
{
	int ret = -1, i;
	int count;

	count = SendMessage(combobox, CB_GETCOUNT, 0, 0);
	for (i = 0; i < count; i++)
	{
		if (SendMessage(combobox, CB_GETITEMDATA, i, 0) == data)
		{
			ret = i;
			break;
		}
	}

	return ret;
}

int OnCBL_SELBYID(HWND combobox, int x)
{
	int ret = -1, i;
	int count;

	count = SendMessage(combobox, CB_GETCOUNT, 0, 0);
	for (i = 0; i < count; i++)
	{
		if (((Link*)SendMessage(combobox, CB_GETITEMDATA, i, 0))->id == x)
		{
			SendMessage(combobox, CB_SETCURSEL, i, 0);
			ret = i;
			break;
		}
	}

	return ret;
}

void OnWM_CREATE(HWND combobox, CREATESTRUCT *cs)
{
	LONG_PTR userdata = 0;

	if (cs->style & LCBS_ALLOWNOSELECT)
	{
		userdata = LCBF_ALLOWNOSELECT;
		cs->style &= ~LCBS_ALLOWNOSELECT;	//don't confuse the ComboWndProc
	}

	SetWindowLongPtr(combobox, GWLP_USERDATA, userdata);
}

LRESULT CALLBACK LComboWndProc(HWND control, UINT msg, WPARAM wParam, LPARAM lParam)
{
	LRESULT ret = 0;

	switch (msg)
	{
	case CBL_FILL:
		ret = OnCBL_FILL(control, (Link*)lParam, (Link*)wParam);
		break;

	case CBL_FILLBYID:
		ret = OnCBL_FILLBYID(control, (Link*)lParam, wParam);
		break;


	case CBL_GETSELID:
		ret = OnCBL_GETSELID(control);
		break;

	case CBL_GETSELPTR:
		ret = (LRESULT)OnCBL_GETSELPTR(control);
		break;

	case CBL_FIND:
		ret = OnCBL_FIND(control, lParam);
		break;

	case CBL_SELBYID:
		ret = OnCBL_SELBYID(control, wParam);
		break;

	case WM_CREATE:
		OnWM_CREATE(control, (CREATESTRUCT*)lParam);

	default:
		if (ComboWndProc)
			ret = CallWindowProc(ComboWndProc, control, msg, wParam, lParam);
	}

	return ret;
}

/* Special subclassed edit control for Tree-Views */
LRESULT CALLBACK TVEditWndProc(HWND control, UINT msg, WPARAM wParam, LPARAM lParam)
{
	LRESULT ret = 0;

	switch (msg)
	{
	case WM_GETDLGCODE:
		ret = DLGC_WANTALLKEYS;
		break;

	default:
		if (EditWndProc)
			ret = CallWindowProc(EditWndProc, control, msg, wParam, lParam);
	}

	return ret;
}

/* Interface */

void RegisterControls(HINSTANCE instance)
{
	WNDCLASS wc;

	GetClassInfo(NULL, "COMBOBOX", &wc);

	ComboWndProc = wc.lpfnWndProc;
	wc.lpfnWndProc = LComboWndProc;
	wc.lpszClassName = LCOMBO;
	wc.hInstance = instance;

	RegisterClass(&wc);

	GetClassInfo(NULL, "EDIT", &wc);

	EditWndProc = wc.lpfnWndProc;
}

void GetWindowText(HWND wnd, SString &value)
{
	int len = GetWindowTextLength(wnd);
	if (len)
	{
		char *cstr = value.unlock(++len);
		GetWindowText(wnd, cstr, len);
		value.lock();
	}
	else
		value.erase();
}

float GetDlgItemFloat(HWND dialog, UINT id)
{
	char buffer[0x20];
	GetDlgItemText(dialog, id, buffer, sizeof(buffer));
	return (float)atof(buffer);
}

void SetDlgItemFloat(HWND dialog, UINT id, float value)
{
	char buffer[0x20];
	_snprintf(buffer, sizeof(buffer), "%.1f", value);
	SetDlgItemText(dialog, id, buffer);
}

UINT GetCheckedRadio(HWND dialog, UINT first, UINT last)
{
	while (first <= last)
	{
		if (SendDlgItemMessage(dialog, first, BM_GETCHECK, 0, 0))
			return first;
		first++;
	}
	return 0;
}

void FillCB(HWND combobox, unsigned count, const struct PAIR *pairs, unsigned select)
{
	unsigned index, type;

	while (count--)
	{
		type = pairs->index;
		index = SendMessage(combobox, CB_ADDSTRING, 0, (LPARAM)pairs->name);
		SendMessage(combobox, CB_SETITEMDATA, index, type);
		if (type == select)
			SendMessage(combobox, CB_SETCURSEL, index, 0);

		pairs++;
	}
}

LPARAM GetItemParam(HWND treeview, HTREEITEM which)
{
	TVITEM item;

	item.mask = TVIF_HANDLE | TVIF_PARAM;
	item.hItem = which ? which : TreeView_GetNextItem(treeview, NULL, TVGN_CARET);
	item.lParam = -1;

	if (item.hItem)	//TVM_GETITIEM does NOT have NULL-protection
		SendMessage(treeview, TVM_GETITEM, 0, (LPARAM)&item);

	return item.lParam;
}

void TreeView_AddChild(HWND treeview, LPARAM param, HTREEITEM parent, HTREEITEM after)
{
	//variables
	HTREEITEM newitem;
	TVINSERTSTRUCT tvis;

	if (!IS_ROOTNODE(treeview, parent))
		parent = TreeView_GetParent(treeview, parent);

	//common settings
	tvis.hParent = parent;
	tvis.hInsertAfter = after;
	tvis.item.mask = TVIF_TEXT | TVIF_PARAM;
	tvis.item.pszText = (LPSTR)LPSTR_TEXTCALLBACK;
	tvis.item.lParam = param;

	//insert the item
	newitem = TreeView_InsertItem(treeview, &tvis);
	SendMessage(treeview, TVM_EXPAND, TVE_EXPAND, (LPARAM)parent);
	TreeView_SelectItem(treeview, newitem);
	SetFocus(treeview);
}

/* Tooltips */

void TooltipInit(HWND tt)
{
	tooltip = tt;
}

BOOL AddTool(HWND dialog, int ctrlId, LPTSTR text)
{
	TOOLINFO ti = { sizeof(TOOLINFO) };

	ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
	ti.hwnd = dialog;
	ti.uId = (UINT_PTR)GetDlgItem(dialog, ctrlId);
	ti.lpszText = text;

	return SendMessage(tooltip, TTM_ADDTOOL, 0, (LPARAM)&ti);
}

